

#include <windows.h>
#include <stdio.h>		// printf() / scanf()
#define _USE_MATH_DEFINES
#include <math.h>
#include <limits.h>

#include <io.h>
#include <string>
#include <vector>
#include <map>

#include "filter.h"
#include "potrace.h"


// [min, max] ͈̔͂ɖOa
template <class T>
T saturate(T v, T min, T max) {
	if (v < min) {
		return min;
	} else if (v > max) {
		return max;
	}
	return v;
}

template <>
int saturate(int v, int min, int max) {
    v -= min;
    v &= ~(v >> 31);
    v += min;

    v -= max;
    v &=  (v >> 31);
    v += max;
    
    return v;
}


static inline int modulo(int a, int b) {
	int m = a % b;

	if (m >= 0) {
		return m;
	} else {
		return m + b;
	}
}



void CopyBufToDIB32(
		PIXEL_YC *buf, DWORD *img,
		int w, int h, int mw, int mh
){
	int src,dst,ix,iy;

	for(iy = 0; iy < h; iy++) {
		src = iy * mw;
		dst = ((h-1)-iy) * w;

		for(ix = 0; ix < w; ix++) {
			int y  = buf[src].y;
			int cb = buf[src].cb;
			int cr = buf[src].cr;

			int r = ( 255*y + ( ((22881*cr>>16)+3) <<10) ) >>12;
			int g = ( 255*y + ( ((-5616*cb>>16)+(-11655*cr>>16)+3) <<10) ) >>12;
			int b = ( 255*y + ( ((28919*cb>>16)+3) <<10) ) >>12;

			r = saturate(r, 0, 255);
			g = saturate(g, 0, 255);
			b = saturate(b, 0, 255);

			img[dst] = (r << 16) + (g << 8) + b;

			src++;
			dst++;
		}
	}
}

static char *g_dllpath = NULL;
static std::map<std::string, int> g_trackMap;
static std::map<std::string, int> g_checkMap;
static std::vector<int> g_bef_check;
static int g_video_rate  = 30;
static int g_video_scale = 1;

static int getTrackVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_trackMap.find(str);

	if (itr == g_trackMap.end()) {
		sprintf_s(errmsg, "track %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	int val = fp->track[itr->second];
	val = std::max<int>(val, fp->track_s[itr->second]);
	val = std::min<int>(val, fp->track_e[itr->second]);

	return val;
}

static int getCheckVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_checkMap.find(str);

	if (itr == g_checkMap.end()) {
		sprintf_s(errmsg, "check %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	return fp->check[itr->second];
}

void initFilter(char *dllpath, FILTER_DLL *fdp) {
	g_dllpath = dllpath;

	g_trackMap.clear();
	for (int i=0; i<fdp->track_n; i++) {
		g_trackMap[fdp->track_name[i]] = i;
	}

	g_checkMap.clear();
	for (int i=0; i<fdp->check_n; i++) {
		g_checkMap[fdp->check_name[i]] = i;
	}
	g_bef_check.resize(fdp->check_n, 0);
}

void updateFilter(FILTER *fp) {
	//fp->exfunc->filter_window_update(fp);
	return;
}


void exitFilter() {
}


#define POTRACE_IN_FN  "__tmp_potrace_in__.bmp"
#define POTRACE_OUT_FN "__tmp_potrace_out__.pgm"


static BOOL saveBitmap(PIXEL_YC *buf, int w, int h, int mw, int mh) {
	BITMAPFILEHEADER bmpFileHeader = {0};
	BITMAPINFOHEADER bmpInfoHeader = {0};

	DWORD* pDIB32 = new DWORD[w*h];
	CopyBufToDIB32(buf, pDIB32, w, h, mw, mh);

	DWORD dwWidth  = w;
	DWORD dwHeight = h;
	DWORD dwLength = w * 4;

	bmpFileHeader.bfSize = sizeof(BITMAPFILEHEADER);
	memcpy(&(bmpFileHeader.bfType), "BM", 2);
	bmpFileHeader.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + dwLength*dwHeight;
	bmpFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

	bmpInfoHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmpInfoHeader.biWidth = dwWidth;
	bmpInfoHeader.biHeight = dwHeight;
	bmpInfoHeader.biPlanes = 1;
	bmpInfoHeader.biBitCount = 32;

	FILE *filep;
	errno_t err = fopen_s(&filep, POTRACE_IN_FN, "wb");
	if (err != 0) {
		delete[] pDIB32;
		return FALSE;
	}
	fwrite(&bmpFileHeader, sizeof(bmpFileHeader), 1, filep);
	fwrite(&bmpInfoHeader, sizeof(bmpInfoHeader), 1, filep);
	fwrite(pDIB32, dwLength*dwHeight, 1, filep);
	fclose(filep);

	delete[] pDIB32;

	return TRUE;
}

class file_p {
public:
	file_p(){
		fp = NULL;
	}

	~file_p() {
		if (fp != NULL) {
			fclose(fp);
		}
	}

	bool open(char *filename, char *mode) {
		errno_t err = fopen_s(&fp, filename, mode);
		if (err != 0) {
			return false;
		}
		return true;
	}

	FILE *p() {
		return this->fp;
	}

private:
	FILE *fp;
};

static BOOL loadPGM(FILTER_PROC_INFO *fpip, PIXEL_YC *buf, int w, int h, int mw, int mh) {
	file_p filep;
	if (!filep.open(POTRACE_OUT_FN, "rb")) {
		return FALSE;
	}

	char sbuf[1024];
	//
	fgets(sbuf, sizeof(sbuf), filep.p());
	if (sbuf[0] != 'P' || sbuf[1] != '5') {
		return FALSE;
	}

	//Rg
	do {
		fgets(sbuf, sizeof(sbuf), filep.p());	
	} while(sbuf[0] == '#');

	//TCY,őPx
	int gw, gh;
	int bmax;
	sscanf_s(sbuf, "%d%d%d", &gw, &gh, &bmax);

	if (gw < 1 || gh < 1) {
		return FALSE;
	}

	BYTE *pSrc = new BYTE[gw*gh];
	fread(pSrc, gw*gh, 1, filep.p());

	int ix,iy;

	for (iy=0; iy<gh && iy<mh; iy++) {
		int src = iy*gw;
		int dst = iy*mw;
		BYTE *srcp = pSrc + src;
		PIXEL_YC *dstp = buf + dst;

		for (ix=0; ix<gw && ix<mw; ix++) {
			PIXEL_YC pix;
			pix.y = *srcp * 16;
			pix.cb = 0;
			pix.cr = 0;

			*dstp = pix;

			srcp++;
			dstp++;
		}
	}

	delete[] pSrc;

	int aw = std::min<int>(gw, mw);
	int ah = std::min<int>(gh, mh);
	fpip->w = aw;
	fpip->h = ah;

	return TRUE;
}


BOOL execFilter(FILTER *fp,FILTER_PROC_INFO *fpip) {
	PIXEL_YC *buffer0 = fpip->ycp_edit;
	PIXEL_YC *buffer1 = fpip->ycp_temp;

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;
	int f = fpip->frame;

	double scale = getTrackVal(fp, "scale") / 100.0;
	int turdsize = getTrackVal(fp, "turdsize");
	double alphamax = getTrackVal(fp, "alphamax") / 10.0;
	double tolerance = getTrackVal(fp, "tolerance") / 10.0;

	//
	BOOL b = saveBitmap(buffer0, w, h, mw, mh);
	if (b != TRUE) {
		return FALSE;
	}

	//potrace
	PROCESS_INFORMATION pi = {0};
	STARTUPINFO si = {0};
	si.cb = sizeof(si);
	si.dwFlags = STARTF_USESHOWWINDOW;
	si.wShowWindow = SW_HIDE;

	TCHAR cmd[2048];
	sprintf_s(cmd, sizeof(cmd), 
		"potrace --scale %4.2f --turdsize %d --alphamax %2.1f --opttolerance %2.1f "POTRACE_IN_FN" -g -o "POTRACE_OUT_FN,
		scale, turdsize, alphamax, tolerance
		);

	{
/*
		FILE *fp;
		fopen_s(&fp, "__potrace_cmd.txt", "w");
		fprintf_s(fp, "%s", cmd);
		fclose(fp);
/**/
	}

	CreateProcess(NULL, cmd, NULL, NULL, FALSE, NORMAL_PRIORITY_CLASS, NULL, NULL, &si, &pi);

	CloseHandle(pi.hThread);

	WaitForSingleObject(pi.hProcess, 1000*10);

	CloseHandle(pi.hProcess);

	//
	loadPGM(fpip, buffer1, w,h, mw,mh);

	fpip->ycp_edit = buffer1;
	fpip->ycp_temp = buffer0;

	return TRUE;
}
